/* Widgets.pmod
 *
 * $Id: Widgets.pmod,v 1.13 1998/04/06 15:48:01 nisse Exp $
 *
 * This file contains generic widget functionality. It should *not*
 * depend on X, instead it communicates with the device object to
 * actually draw anything, or arrange for callbacks.
 */

/*
 *    px, a Pike interface to the X Window System
 *
 *    Copyright (C) 1998, Niels Mller, Per Hedbor, Marcus Comstedt,
 *    Pontus Hagland, David Hedbor.
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */

/* Questions, bug fixes and bug reports can be sent to the pike
 * mailing list, pike@idonex.se, or to the athors (see AUTHORS for
 * email addresses. */

#include "error.h"

class Widget
{
  /* Output device for the widget to interact with */
  object device;

  /* Parent widget */
  object parent;
  
  int x, y;  // Relative to the parent widget 
  int width, height;
  
  /* Events are identified by name.*/
  mapping(string:function) callbacks = ([ ]);

  /* Various data associated with the gadget.
   * Keys beginning with _ are reserved for internal and device use. */
   
  mapping(string:mixed) data = ([ ]);

  /* This function creates somewhat non-trivial interactions
   * between the widget tree and the inheritance tree... */
  mixed signal_this(string name, mixed ...args)
  {
    werror(sprintf("signal_this('%s')\n"
		   "callbacks = %O\n", name, callbacks));
    function f = (!zero_type(callbacks[name]))
      ? callbacks[name] : this_object()[name];
    return f && f(@args);
  }

  mixed signal_recurse(string name, mixed ...args)
  {
    return signal_this(name, @args);
  }
  
  mixed signal(string name, mixed ...args)
  {
    return signal_this(name, @args) || (parent && parent->signal(name, @args));
  }
  
  /* The widget should draw itself with its upper left corner at (x,
   * y) on the parent widget. */
  
  // FIXME: Is this method really needed? Aren't show() and refresh() enough?
  // It's probably useful, when a window is redrawn after resize,
  // it shouldn't be called by the user of a widget.
#if 0
  void draw()
  {
  }
#endif

  /* Returns the widget's preferred width and height. */
  mapping query_geometry()
  {
    return 0;
  }

  void set_geometry(mapping m)
  {
    x = m->x;
    y = m->y;
    width = m->width;
    height = m->height;
  }

#if 0
  void show()
  {
  }
#endif
  
  mixed query_data(string s)
  {
    return data[s] || (parent && parent->query_data(s));
  }

  /* Returns an array ({ context, x, y }),
   * where context is the most recent device context, and
   * x, y are the offset to the origin of this widget.
   *
   * if (super), ignore context associated with this widget,
   * and start the search in the parent.
  
  /* FIXME: This data could be cached, but it must be recalculated
   * occasionally, e.g. when a window is resized */
  array query_device_context(int|void super)
  {
    if (!super && data->_context)
      return ({ data->_context, data->_context->x, data->_context->y });
    array a = parent->query_device_context();
    if (a)
      {
	a[1] += x; a[2] += y;
      }
    return a;
  }
  
  void use_device(object d)
  {
    device = d;
  }

  /* Redrawing the widget... */

  /* Called by the device when the widget needs to be redrawn */
  // FIXME: This function should get some (abstract) exposure event as argument
#if 0
  void refresh(int x, int y, int width, int height, int last)
  {
    if (last)
      return signal_this("draw");
    return 0;
  }
#endif
  
  /* Called by parent widget if it thinks that the widget needs to be redrawn.
   * Note that for some devices some widgets (i.e. X buttons) may have their own
   * exposure callbacks, and should ignore the parent calling this function. */
  // FIXME: This function should get some (abstract) exposure event as argument
  void refresh_parent(int ex, int ey, int width, int height, int last)
  {
    if (!data->_ignore_parent_refresh)
      signal_this("refresh", ex - x, ey - y, width, height, last);
  }

  mixed exposed(int x, int y, int width, int height, int last)
  {
    return signal_recurse("refresh", x, y, width, height, last);
  }
}

class default_refresh
{
  // mixed signal_this(string name, mixed ...args);
  
  void refresh(int x, int y, int width, int height, int last)
  {
    if (last)
      return this_object()->signal_this("draw");
    return 0;
  }
}

class Container
{
  inherit Widget;

  array children = ({ });

  mixed signal_recurse(string name, mixed ...args)
  {
    werror(sprintf("Widget.Container->signal_recurse(%s)\n", name));
    string recurse = (name == "refresh")
      ? "refresh_parent" : name;
    return signal_this(name, @args) || children->signal_recurse(recurse, @args);
  }
  
  void use_device(object d)
  {
    Widget::use_device(d);
    children->use_device(d);
  }
  
  void add(object o)
  {
    children += ({ o });
    o->parent = this_object();
  }
}

class Bin
{
  inherit Widget;
  object child;

  mixed signal_recurse(string name, mixed ...args)
  {
    string recurse = (name == "refresh")
      ? "refresh_parent" : name;
    return signal_this(name, @args) || child->signal_recurse(recurse, @args);
  }

  mapping query_geometry()
  {
    return child->query_geometry();
  }

  void set_geometry(mapping m)
  {
    m = copy_value(m);
    m->x = m->y = 0; // Relative to this widget
    child->set_geometry(m);
  }

  void set_my_geometry(mapping m)
  {
    Widget::set_geometry(m);
  }
  
  void use_device(object d)
  {
    Widget::use_device(d);
    child->use_device(d);
  }

  object add(object o)
  {
    if (child)
      error("Widget.Bin->add: Already full!\n");
    child = o;
    o->parent = this_object();
  }
}


class Button
{
  inherit Bin;

  int button_nr = 0;
  
  mixed draw()
  {
    device->draw_button(this_object());
    // werror("Widget.Button->draw: not implemented.\n");
    return child->signal_this("draw");
  }

  mapping query_geometry()
  {
    return device->button_geometry(child->query_geometry());
  }

  void set_geometry(mapping m)
  {
    werror(sprintf("Button->set_geometry: %O\n", m));
    set_my_geometry(m);
    child->set_geometry(device->button_interior_geometry(m));
    int w;
    device->set_button_geometry(this_object());
    {
      // werror(sprintf("Button->set_geometry: window = %O\n",
      //  indices(data->window)));
    }
  }

  void set_clicked_callback(function f)
  {
    callbacks->button_clicked = f;
  }
  
  mixed show()
  {
    device->create_button(this_object());
    // device->set_button_callback(this_object(), clicked_callback);
    child->signal_recurse("show");
    device->show_button(this_object());
    return 1;
  }
}

// FIXME: How should backgrounds be handled?
class Window
{
  inherit Bin;

  string title;
  string wmclass_class; // These properties are rather X specific
  string wmclass_name;

  int min_width;
  int min_height;
  
  mixed resized(int w, int h)
  {
    if ((w != width) || (h != height))
      {
	werror("Widget.Window->resized\n");
	mapping m = ([ "width" : w,
		       "height" : h,
		       "x" : 0, "y" : 0]);
	set_my_geometry(m);
	child->set_geometry(m);
	device->clear_window(this_object());
	signal_this("draw");
      }
    return 1;
  }
  
  mixed show()
  {
    mapping m = child->query_geometry();

    min_width = m->width;
    min_height = m->height;
    
    if (width < min_width)
      width = min_width;
      
    if (height < min_height)
      height = min_height;

    child->set_geometry(m | ([ "x" :0, "y" : 0 ]) );

    device->create_window(this_object());
    // device->set_resize_callback(this_object(), resized);
    child->signal_recurse("show");
    device->show_window(this_object());
    return 1;
  }
}

class Box
{
  inherit Container;

  /* Apply constraints, increasing the size of some elements */
  array(int) apply_constraints(array(int) sizes,
			       array(multiset(int)) constraints)
  {
    array(int) res = copy_value(sizes);
    foreach(constraints, multiset set)
      {
	int size = sort(rows(sizes, indices(set)))[-1];
	foreach(indices(set), int i)
	  res[i] = size;
      }
    return res;
  }

  /* Stretch */
  array(int) apply_stretch(array(int) sizes, array(int) stretch, int real_size)
  {
    int extra_space = real_size - `+( @ sizes );
    int sum_stretch = stretch ? `+( @ stretch ) : sizeof(sizes);
    
    if (extra_space > 0)
      {
	// werror(sprintf("Box->apply_stretch: sizes = %O\n"
	// 		  "extra_space = %d, sum_stretch = %d\n",
	// 		  sizes, extra_space, sum_stretch));
	array(int) res = copy_value(sizes);
	for(int i; i<sizeof(sizes); i++)
	  {
	    int s = stretch ? stretch[i] : 1;
	    int expand = s * extra_space / sum_stretch;
	    res[i] += expand;
	    extra_space -= expand;
	    sum_stretch -= s;
	    // werror(sprintf("Box->apply_stretch: s = %d, expand = %d,\n"
	    // 		      "extra_space = %d, sum_stretch  = %d\n"
	    // 		      "res = %O\n",
	    // 	      s, expand, extra_space, sum_stretch, res));
	  }
	return res;
      }
    return sizes;
  }
}

class HBox
{
  inherit Box;
  array(multiset(int)) constraints;
  array(int) stretch;

  array(int) preferred_widths; // Allocated widths for the children
  int preferred_height;
  
  int spacing;
  
  mapping query_geometry()
  {
    array gs = children->query_geometry();
    mapping m = ([ ]);
    preferred_height = sort(gs->height)[-1]; // max;
    preferred_widths = gs->width;
    if (constraints)
      preferred_widths
	= apply_constraints(preferred_widths, constraints);
    return ([ "width" : `+(@ preferred_widths)
                        + spacing * (sizeof(children) - 1),
	      "height" : preferred_height ]);
  }

  void set_geometry(mapping m)
  {
    // trace(1);
    Box::set_geometry(m);
    int space = m->width - spacing * (sizeof(children) - 1);

    array(int) alloc = apply_stretch(preferred_widths, stretch, space);

    int pos = 0;
    for (int i=0; i<sizeof(children); i++)
      {
	children[i]->set_geometry( ([ "width" : alloc[i],
				      "height" : m->height,
				      "x": pos, "y" : 0]) );
	pos += alloc[i] + spacing;
      }
    // trace(0);
  }
}

class VBox
{
  inherit Box;
  array(multiset(int)) constraints;
  array(int) stretch;

  array(int) preferred_heights; // Allocated heights for the children
  int preferred_width;
  
  int spacing;
  
  mapping query_geometry()
  {
    array gs = children->query_geometry();
    mapping m = ([ ]);
    preferred_width = sort(gs->width)[-1]; // max;
    preferred_heights = gs->height;
    if (constraints)
      preferred_heights
	= apply_constraints(preferred_heights, constraints);
    return ([ "width" : preferred_width,
	      "height" : `+(@ preferred_heights)
                        + spacing * (sizeof(children) - 1) ]);
  }

  void set_geometry(mapping m)
  {
    Box::set_geometry(m);
    int space = m->height - spacing * (sizeof(children) - 1);

    array(int) alloc = apply_stretch(preferred_heights, stretch, space);

    int pos = 0;
    for (int i=0; i< sizeof(children); i++)
      {
	children[i]->set_geometry( ([ "width" : m->width,
				      "height" :  alloc[i],
				      "x": 0, "y" : pos]) );
	pos += alloc[i] + spacing;
      }
  }
} 

class Table
{
  inherit Box;
  int rows, columns;
  int index; // Used by add()
  
  array(multiset(int)) x_constraints;
  array(multiset(int)) y_constraints;

  array(int) x_stretch;
  array(int) y_stretch;
  
  array(int) preferred_widths;
  array(int) preferred_heights;

  int x_spacing, y_spacing;

  void create(int r, int c)
  {
    rows = r;
    columns = c;
    children = allocate(rows*columns);
    index = 0;
  }

  void add(object widget)
  {
    children[index++] = widget;
    widget->parent = this_object();
  }

  void set(object widget, int r, int c)
  {
    children[r*columns + c] = widget;
    widget->parent = this_object();
  }

  object get(int r, int c)
  {
    return children[r*columns + c];
  }

  array(object) get_row(int r, array|void a)
  {
    return (a||children)[r*columns .. r*columns+columns-1];
  }

  array(object) get_column(int c, array|void a)
  {
    array res = allocate(rows);
    
    for (int i=0; i<columns; i++, c += columns)
      res[i] = (a||children)[c];
    return res;
  }
  
  mapping query_geometry()
  {
    array gs = children->query_geometry();
    preferred_widths = allocate(columns);
    preferred_heights = allocate(rows);
    
    int i;
    
    for(i=0; i<columns; i++)
      preferred_widths[i] = sort(get_column(i, gs)->width)[-1]; // max
    if (x_constraints)
      preferred_widths 
	= apply_constraints(preferred_widths, x_constraints);

    for(i=0; i<rows; i++)
      preferred_heights[i] = sort(get_row(i, gs)->height)[-1]; // max
    if (y_constraints)
      preferred_heights
	= apply_constraints(preferred_heights, y_constraints);
    
    return ([ "width" : `+(@ preferred_widths)
                        + x_spacing * (columns - 1),
	      "height": `+(@ preferred_heights)
                        + y_spacing * (rows - 1) ]);
  }

  void set_geometry(mapping m)
  {
    Box::set_geometry(m);

    int x_space = m->width - x_spacing * (columns - 1);
    array(int) x_alloc = apply_stretch(preferred_widths, x_stretch, x_space);

    int y_space = m->height - y_spacing * (rows - 1);
    array(int) y_alloc = apply_stretch(preferred_heights, y_stretch, y_space);

    int x_pos, y_pos;
    
    int i = 0;
    int y_pos = 0;
    for(int r = 0; r<rows; r++)
      {
	x_pos = 0;
	for(int c = 0; c<columns; c++)
	  {
	    children[i++]->set_geometry( ([ "width" : x_alloc[c],
					    "height" : y_alloc[r],
					    "x" : x_pos,
					    "y" : y_pos ]) );
	    x_pos += x_alloc[c] + x_spacing;
	  }
	y_pos += y_alloc[r] + y_spacing;
      }
  }
}
  
class Misc
{
  inherit Widget;
}

class Solid
{
  inherit Misc;
  inherit default_refresh;
  
  int preferred_width, preferred_height;
  
  /* Returns the widget's preferred width and height. */
  mapping query_geometry()
  {
    return ([ "width" : preferred_width,
	    "height" : preferred_height ]);
  }

#if 0
  // DEBUG
  void set_geometry(mapping m)
  {
    werror(sprintf("Solid->set_geometry: %O\n", m));
    Misc::set_geometry(m);
  }
#endif
  
  mixed draw()
  {
    device->draw_solid(this_object());
    return 1;
  }
}

class Image
{
  inherit Misc;

  object image;
  
  mapping query_geometry()
  {
    werror(sprintf("Widget.Image->query_geometry: %d, %d\n",
	   image->xsize(), image->ysize()));
    return ([ "width" : image->xsize(), "height" : image->ysize() ]);
  }

  void set_geometry(mapping m)
  {
    Misc::set_geometry(m);
    device->set_image_geometry(this_object());
  }
  
  mixed show()
  {
    image = device->create_image(this_object(), image);
    device->show_image(this_object());
    return 1;
  }
  
  mixed refresh(int x, int y, int width, int height, int last)
  {
    werror(sprintf("Widget.Image->refresh: %d, %d, %d, %d, %d\n",
		   x, y, width, height, last));
    device->draw_image(this_object(), x, y, width, height);
    return 1;
  }

  mixed draw()
  {
    device->draw_image(this_object(), 0, 0, width, height);
    return 1;
  }
}

class Alignment
{
  inherit Bin;

  float xscale, yscale;
  float xalign, yalign;

  array align(int requested, int assigned, float scale, float align)
  {
    int diff = assigned - requested;
    if (diff > 0)
      {
	int expansion = (int) (scale * diff);
	return ({ expansion, (int) (align * (diff - expansion)) });
      }
    else
      return ({ 0, 0 });
  }
  
  void set_geometry(mapping m)
  {
    werror(sprintf("Alignment->set_geometry: %O\n", m));
    set_my_geometry(m);

    mapping geometry = child->query_geometry();
    mapping aligned = ([ ]); 

    array a;

    a = align(geometry->width, m->width, xscale, xalign);
    aligned->width = geometry->width + a[0];
    aligned->x = a[1];
    
    a = align(geometry->height, m->height, yscale, yalign);
    aligned->height = geometry->height + a[0];
    aligned->y = a[1];

    child->set_geometry(aligned);
  }
}

